from pymoo.algorithms.nsga3 import NSGA3
from pymoo.factory import get_problem, get_reference_directions
from pymoo.optimize import minimize
from pymoo.visualization.scatter import Scatter

problem = get_problem("DASCMOP1", 2)

ref_dirs = get_reference_directions("das-dennis", problem.n_obj, n_points=91)
algorithm = NSGA3(ref_dirs)

res = minimize(problem,
               algorithm,
               ("n_gen", 600),
               verbose=True
               )

plot = Scatter()
plot.add(problem.pareto_front(), color="black", alpha=0.7)
plot.add(res.F, color="red")
plot.show()



from pymoo.algorithms.nsga3 import NSGA3
from pymoo.factory import get_reference_directions
from pymoo.optimize import minimize
from pymoo.problems.multi.omnitest import OmniTest
from pymoo.visualization.scatter import Scatter
import matplotlib.pyplot as plt

problem = OmniTest(n_var=3)
ref_dirs = get_reference_directions("das-dennis", problem.n_obj, n_partitions=30)
PS = problem.pareto_set(5000)
PF = problem.pareto_front(5000)

algorithm = NSGA3(ref_dirs=ref_dirs)

res = minimize(problem,
               algorithm,
               ('n_gen', 500),
               seed=1,
               verbose=False)

fig_name = f"{algorithm.__class__.__name__} on Omni-test"

# visualize decision space
plot = Scatter(title="Decision Space")
plot.add(PS, s=10, color='r', label="PS")
plot.add(res.X, s=30, color='b', label="Obtained solutions")
plot.do()
plt.legend()

# visualize objective space
plot = Scatter(title="Objective Space")
plot.add(PF, s=10, color='r', label="PF")
plot.add(res.F, s=30, color='b', label="Obtained solutions")
plot.do()
plt.legend()

plt.show()




from pymoo.algorithms.so_direct import DIRECT
from pymoo.optimize import minimize
from pymoo.problems.single import Himmelblau

problem = Himmelblau()

algorithm = DIRECT()

ret = minimize(problem,
               algorithm,
               verbose=True)




import numpy as np

from pymoo.algorithms.so_genetic_algorithm import FitnessSurvival
from pymoo.docs import parse_doc_string
from pymoo.model.algorithm import Algorithm, filter_optimum
from pymoo.model.initialization import Initialization
from pymoo.model.population import Population
from pymoo.model.replacement import ImprovementReplacement, is_better
from pymoo.operators.mutation.polynomial_mutation import PolynomialMutation
from pymoo.operators.repair.inverse_penalty import InversePenaltyOutOfBoundsRepair
from pymoo.operators.repair.to_bound import set_to_bounds_if_outside
from pymoo.operators.sampling.latin_hypercube_sampling import LatinHypercubeSampling
from pymoo.util.display import SingleObjectiveDisplay
from pymoo.util.misc import norm_eucl_dist, norm_euclidean_distance
from pymoo.util.termination.default import SingleObjectiveDefaultTermination
from pymoo.visualization.fitness_landscape import FitnessLandscape
from pymoo.visualization.video.callback_video import AnimationCallback


# =========================================================================================================
# Display
# =========================================================================================================

class PSODisplay(SingleObjectiveDisplay):

    def _do(self, problem, evaluator, algorithm):
        pop = algorithm.pop
        algorithm.pop = Population.create(*algorithm.pop.get("pbest"))
        super()._do(problem, evaluator, algorithm)
        algorithm.pop = pop

        if algorithm.adaptive:
            self.output.append("f", algorithm.f if algorithm.f is not None else "-", width=8)
            self.output.append("S", algorithm.strategy if algorithm.strategy is not None else "-", width=6)
            self.output.append("w", algorithm.w, width=6)
            self.output.append("c1", algorithm.c1, width=8)
            self.output.append("c2", algorithm.c2, width=8)


# =========================================================================================================
# Adaptation Constants
# =========================================================================================================


def S1_exploration(f):
    if f <= 0.4:
        return 0
    elif 0.4 < f <= 0.6:
        return 5 * f - 2
    elif 0.6 < f <= 0.7:
        return 1
    elif 0.7 < f <= 0.8:
        return -10 * f + 8
    elif 0.8 < f:
        return 0


def S2_exploitation(f):
    if f <= 0.2:
        return 0
    elif 0.2 < f <= 0.3:
        return 10 * f - 2
    elif 0.3 < f <= 0.4:
        return 1
    elif 0.4 < f <= 0.6:
        return -5 * f + 3
    elif 0.6 < f:
        return 0


def S3_convergence(f):
    if f <= 0.1:
        return 1
    elif 0.1 < f <= 0.3:
        return -5 * f + 1.5
    elif 0.3 < f:
        return 0


def S4_jumping_out(f):
    if f <= 0.7:
        return 0
    elif 0.7 < f <= 0.9:
        return 5 * f - 3.5
    elif 0.9 < f:
        return 1


# =========================================================================================================
# Implementation
# =========================================================================================================


class PSO(Algorithm):

    def __init__(self,
                 pop_size=25,
                 sampling=LatinHypercubeSampling(),
                 w=0.9,
                 c1=2.0,
                 c2=2.0,
                 adaptive=True,
                 initial_velocity="random",
                 max_velocity_rate=0.20,
                 pertube_best=True,
                 display=PSODisplay(),
                 **kwargs):
        """

        Parameters
        ----------
        pop_size : The size of the swarm being used.

        sampling : {sampling}

        adaptive : bool
            Whether w, c1, and c2 are changed dynamically over time. The update uses the spread from the global
            optimum to determine suitable values.

        w : float
            The inertia weight to be used in each iteration for the velocity update. This can be interpreted
            as the momentum term regarding the velocity. If `adaptive=True` this is only the
            initially used value.

        c1 : float
            The cognitive impact (personal best) during the velocity update. If `adaptive=True` this is only the
            initially used value.
        c2 : float
            The social impact (global best) during the velocity update. If `adaptive=True` this is only the
            initially used value.

        initial_velocity : str - ('random', or 'zero')
            How the initial velocity of each particle should be assigned. Either 'random' which creates a
            random velocity vector or 'zero' which makes the particles start to find the direction through the
            velocity update equation.


        max_velocity_rate : float
            The maximum velocity rate. It is determined variable (and not vector) wise. We consider the rate here
            since the value is normalized regarding the `xl` and `xu` defined in the problem.

        pertube_best : bool
            Some studies have proposed to mutate the global best because it has been found to converge better.
            Which means the population size is reduced by one particle and one function evaluation is spend
            additionally to permute the best found solution so far.

        """

        super().__init__(display=display, **kwargs)

        self.initialization = Initialization(sampling)

        self.pop_size = pop_size
        self.adaptive = adaptive
        self.pertube_best = pertube_best
        self.default_termination = SingleObjectiveDefaultTermination()
        self.V_max = None
        self.initial_velocity = initial_velocity
        self.max_velocity_rate = max_velocity_rate

        self.w = w
        self.c1 = c1
        self.c2 = c2

    def setup(self, problem, **kwargs):
        super().setup(problem, **kwargs)
        self.V_max = self.max_velocity_rate * (problem.xu - problem.xl)

    def _initialize(self):
        pop = self.initialization.do(self.problem, self.pop_size, algorithm=self)
        self.evaluator.eval(self.problem, pop, algorithm=self)

        if self.pertube_best:
            pop = FitnessSurvival().do(self.problem, pop, self.pop_size - 1)

        if self.initial_velocity == "random":
            init_V = np.random.random((len(pop), self.problem.n_var)) * self.V_max[None, :]
        elif self.initial_velocity == "zero":
            init_V = np.zeros((len(pop), self.problem.n_var))

        pop.set("V", init_V)
        pop.set("pbest", pop)
        self.pop, self.off = pop, pop

        self.f = None
        self.strategy = None

    def _next(self):
        self._step()

        if self.adaptive:
            self._adapt()

    def _step(self):
        pop = self.pop
        X, F, V = pop.get("X", "F", "V")

        # get the personal best of each particle
        pbest = Population.create(*pop.get("pbest"))
        P_X, P_F = pbest.get("X", "F")

        # get the GLOBAL best solution - other variants such as local best can be implemented here too
        best = self.opt.repeat(len(pop))
        G_X = best.get("X")

        # get the inertia weight of the individual
        inerta = self.w * V

        # calculate random values for the updates
        r1 = np.random.random((len(pop), self.problem.n_var))
        r2 = np.random.random((len(pop), self.problem.n_var))

        cognitive = self.c1 * r1 * (P_X - X)
        social = self.c2 * r2 * (G_X - X)

        # calculate the velocity vector
        _V = inerta + cognitive + social
        _V = set_to_bounds_if_outside(_V, - self.V_max, self.V_max)

        # update the values of each particle
        _X = X + _V
        _X = InversePenaltyOutOfBoundsRepair().do(self.problem, _X, P=X)

        # evaluate the offspring population
        off = Population(len(pop)).set("X", _X, "V", _V, "pbest", pbest)
        self.evaluator.eval(self.problem, off, algorithm=self)

        # check whether a solution has improved or not - also consider constraints here
        has_improved = ImprovementReplacement().do(self.problem, pbest, off, return_indices=True)

        # replace the personal best of each particle if it has improved
        off[has_improved].set("pbest", off[has_improved])
        off.set("best", best)
        pop = off

        # try to improve the current best with a pertubation
        if self.pertube_best:
            pbest = Population.create(*pop.get("pbest"))
            k = FitnessSurvival().do(self.problem, pbest, 1, return_indices=True)[0]
            eta = int(np.random.uniform(5, 30))
            mutant = PolynomialMutation(eta).do(self.problem, pbest[[k]])[0]
            self.evaluator.eval(self.problem, mutant, algorithm=self)

            # if the mutant is in fact better - replace the personal best
            if is_better(mutant, pop[k]):
                pop[k].set("pbest", mutant)

        self.pop = pop

    def _adapt(self):
        pop = self.pop

        X, F, best = pop.get("X", "F", "best")
        best = Population.create(*best)
        w, c1, c2, = self.w, self.c1, self.c2

        # get the average distance from one to another for normalization
        D = norm_eucl_dist(self.problem, X, X)
        mD = D.sum(axis=1) / (len(pop) - 1)
        _min, _max = mD.min(), mD.max()

        # get the average distance to the global best
        g_D = norm_euclidean_distance(self.problem)(best.get("X"), X).mean()
        f = (g_D - _min) / (_max - _min + 1e-32)

        S = np.array([S1_exploration(f), S2_exploitation(f), S3_convergence(f), S4_jumping_out(f)])
        strategy = S.argmax() + 1

        delta = 0.05 + (np.random.random() * 0.05)

        if strategy == 1:
            c1 += delta
            c2 -= delta
        elif strategy == 2:
            c1 += 0.5 * delta
            c2 -= 0.5 * delta
        elif strategy == 3:
            c1 += 0.5 * delta
            c2 += 0.5 * delta
        elif strategy == 4:
            c1 -= delta
            c2 += delta

        c1 = max(1.5, min(2.5, c1))
        c2 = max(1.5, min(2.5, c2))

        if c1 + c2 > 4.0:
            c1 = 4.0 * (c1 / (c1 + c2))
            c2 = 4.0 * (c2 / (c1 + c2))

        w = 1 / (1 + 1.5 * np.exp(-2.6 * f))

        self.f = f
        self.strategy = strategy
        self.c1 = c1
        self.c2 = c2
        self.w = w

    def _set_optimum(self, force=False):
        pbest = Population.create(*self.pop.get("pbest"))
        self.opt = filter_optimum(pbest, least_infeasible=True)


# =========================================================================================================
# Animation
# =========================================================================================================

class PSOAnimation(AnimationCallback):

    def __init__(self,
                 nth_gen=1,
                 n_samples_for_surface=200,
                 dpi=200,
                 **kwargs):
        super().__init__(nth_gen=nth_gen, dpi=dpi, **kwargs)
        self.n_samples_for_surface = n_samples_for_surface
        self.last_pop = None

    def do(self, problem, algorithm):
        import matplotlib.pyplot as plt

        if problem.n_var != 2 or problem.n_obj != 1:
            raise Exception(
                "This visualization can only be used for problems with two variables and one objective!")

        # draw the problem surface
        FitnessLandscape(problem,
                         _type="contour",
                         kwargs_contour=dict(alpha=0.3),
                         n_samples=self.n_samples_for_surface).do()

        # get the population
        off = algorithm.pop
        pop = algorithm.pop if self.last_pop is None else self.last_pop
        pbest = Population.create(*off.get("pbest"))

        for i in range(len(pop)):
            plt.plot([off[i].X[0], pop[i].X[0]], [off[i].X[1], pop[i].X[1]], color="blue", alpha=0.5)
            plt.plot([pbest[i].X[0], pop[i].X[0]], [pbest[i].X[1], pop[i].X[1]], color="red", alpha=0.5)
            plt.plot([pbest[i].X[0], off[i].X[0]], [pbest[i].X[1], off[i].X[1]], color="red", alpha=0.5)

        X, F, CV = pbest.get("X", "F", "CV")
        plt.scatter(X[:, 0], X[:, 1], edgecolors="red", marker="*", s=70, facecolors='none', label="pbest")

        X, F, CV = off.get("X", "F", "CV")
        plt.scatter(X[:, 0], X[:, 1], color="blue", marker="o", s=30, label="particle")

        X, F, CV = pop.get("X", "F", "CV")
        plt.scatter(X[:, 0], X[:, 1], color="blue", marker="o", s=30, alpha=0.5)

        opt = algorithm.opt
        X, F, CV = opt.get("X", "F", "CV")
        plt.scatter(X[:, 0], X[:, 1], color="black", marker="x", s=100, label="gbest")

        xl, xu = problem.bounds()
        plt.xlim(xl[0], xu[0])
        plt.ylim(xl[1], xu[1])

        plt.title(f"Generation: %s \nf: %.5E" % (algorithm.n_gen, opt[0].F[0]))
        plt.legend()

        self.last_pop = off.copy(deep=True)


parse_doc_string(PSO.__init__)



import copy


class Individual:

    def __init__(self, X=None, F=None, CV=None, G=None, feasible=None, **kwargs) -> None:
        self.X = X
        self.F = F
        self.CV = CV
        self.G = G
        self.feasible = feasible
        self.data = kwargs
        self.attr = set(self.__dict__.keys())

    def has(self, key):
        return key in self.attr or key in self.data

    def set(self, key, value):
        if key in self.attr:
            self.__dict__[key] = value
        else:
            self.data[key] = value
        return self

    def copy(self, deep=False):
        ind = copy.copy(self)
        ind.data = copy.copy(self.data) if not deep else copy.deepcopy(self.data)
        return ind

    def get(self, *keys):

        def _get(key):
            if key in self.data:
                return self.data[key]
            elif key in self.attr:
                return self.__dict__[key]
            else:
                return None

        ret = []

        for key in keys:
            ret.append(_get(key))

        if len(ret) == 1:
            return ret[0]
        else:
            return tuple(ret)


class eIndividual:

    def __init__(self, **kwargs) -> None:
        kwargs = {**kwargs, **dict(X=None, F=None, CV=None, G=None, feasible=None)}
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def has(self, key):
        return key in self.__dict__

    def set(self, key, val):
        self.__dict__[key] = val

    def get(self, *keys):
        if len(keys) == 1:
            return self.__dict__.get(keys[0])
        else:
            return tuple([self.__dict__.get(key) for key in keys])

    def copy(self, deep=False):
        ind = copy.copy(self)
        ind.data = copy.copy(self.__dict__) if not deep else copy.deepcopy(self.__dict__)
        return ind


        if not deep:
            d = dict(self.__dict__)
        else:
            d = copy.deepcopy(self.__dict__)
        ind = Individual(**d)
        return ind

    # @property
    # def F(self):
    #     attr = "F"
    #     if attr in self.__dict__:
    #         return self.__dict__[attr]
    #     else:
    #         return None

    # Gets called when the item is not found via __getattribute__
    # def __getattr__(self, item):
    #     return super(Individual, self).__setattr__(item, 'orphan')

    def __getattr__(self, val):
        return self.__dict__.get(val)

    # def __setitem__(self, key, value):
    #     self.__dict__[key] = value
    #
    # def __getitem__(self, key):
    #     return self.__dict__.get(key)

    # def __getattr__(self, attr):
    #
    #     if attr == "F":
    #         if attr in self.__dict__:
    #             return self.__dict__[attr]
    #         else:
    #             return None
    #
    #     if attr in self.__dict__:
    #         return self.__dict__[attr]
    #
    #
    #
    def __setattr__(self, key, value):
        self.__dict__[key] = value



from pymoo.algorithms.so_ga_niching import NicheGA
from pymoo.optimize import minimize
from pymoo.problems.single import Rastrigin
from pymoo.problems.single.simple import SimpleMultiModal01

problem = SimpleMultiModal01()

problem = Rastrigin(n_var=30)

algorithm = NicheGA(pop_size=100)

res = minimize(problem,
               algorithm,
               seed=1,
               verbose=True)


print(res.F)


from pymoo.algorithms.nsga2 import NSGA2
from pymoo.factory import get_problem, get_termination
from pymoo.visualization.scatter import Scatter

problem = get_problem("zdt1")

algorithm = NSGA2(pop_size=100, termination=get_termination('n_gen', 200))

algorithm.setup(problem, verbose=True)

while algorithm.has_next():
    algorithm.next()

res = algorithm.result()

plot = Scatter()
plot.add(problem.pareto_front(), plot_type="line", color="black", alpha=0.7)
plot.add(res.opt.get("F"), color="red")
plot.show()




import numpy as np

from pymoo.model.problem import FunctionalProblem

objs = [
    lambda x: np.sum((x - 2) ** 2),
    lambda x: np.sum((x + 2) ** 2)
]

constr_ieq = [
    lambda x: np.sum((x - 1) ** 2)
]


problem = FunctionalProblem(10,
                            objs,
                            constr_ieq=constr_ieq,
                            xl=np.array([-10, -5, -10]),
                            xu=np.array([10, 5, 10])
                            )

F, CV = problem.evaluate(np.random.rand(3, 10))

print(f"F: {F}\n")
print(f"CV: {CV}")


import matplotlib
matplotlib.use('Agg')

from pymoo.model.problem import Problem

import numpy as np


RESULT = None


class MyProblem(Problem):

    def __init__(self):
        super().__init__(n_var=2,
                         n_obj=2,
                         n_constr=2,
                         xl=np.array([-2, -2]),
                         xu=np.array([2, 2]))

    def _evaluate(self, X, out, *args, **kwargs):
        f1 = X[:, 0] ** 2 + X[:, 1] ** 2
        f2 = (X[:, 0] - 1) ** 2 + X[:, 1] ** 2

        g1 = 2 * (X[:, 0] - 0.1) * (X[:, 0] - 0.9) / 0.18
        g2 = - 20 * (X[:, 0] - 0.4) * (X[:, 0] - 0.6) / 4.8

        out["F"] = np.column_stack([f1, f2])
        out["G"] = np.column_stack([g1, g2])


vectorized_problem = MyProblem()

import numpy as np
from pymoo.model.problem import Problem


class MyProblem(Problem):

    def __init__(self):
        super().__init__(n_var=2,
                         n_obj=2,
                         n_constr=2,
                         xl=np.array([-2, -2]),
                         xu=np.array([2, 2]),
                         elementwise_evaluation=True)

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = x[0] ** 2 + x[1] ** 2
        f2 = (x[0] - 1) ** 2 + x[1] ** 2

        g1 = 2 * (x[0] - 0.1) * (x[0] - 0.9) / 0.18
        g2 = - 20 * (x[0] - 0.4) * (x[0] - 0.6) / 4.8

        out["F"] = [f1, f2]
        out["G"] = [g1, g2]


elementwise_problem = MyProblem()


from pymoo.model.problem import FunctionalProblem

objs = [
    lambda x: x[0]**2 + x[1]**2,
    lambda x: (x[0]-1)**2 + x[1]**2
]

constr_ieq = [
    lambda x: 2*(x[0]-0.1) * (x[0]-0.9) / 0.18,
    lambda x: - 20*(x[0]-0.4) * (x[0]-0.6) / 4.8
]

functional_problem = FunctionalProblem(2,
                                       objs,
                                       constr_ieq=constr_ieq,
                                       xl=np.array([-2,-2]),
                                       xu=np.array([2,2]))


def func_pf(flatten=True, **kwargs):
    f1_a = np.linspace(0.1 ** 2, 0.4 ** 2, 100)
    f2_a = (np.sqrt(f1_a) - 1) ** 2

    f1_b = np.linspace(0.6 ** 2, 0.9 ** 2, 100)
    f2_b = (np.sqrt(f1_b) - 1) ** 2

    a, b = np.column_stack([f1_a, f2_a]), np.column_stack([f1_b, f2_b])
    return stack(a, b, flatten=flatten)


def func_ps(flatten=True, **kwargs):
    x1_a = np.linspace(0.1, 0.4, 50)
    x1_b = np.linspace(0.6, 0.9, 50)
    x2 = np.zeros(50)

    a, b = np.column_stack([x1_a, x2]), np.column_stack([x1_b, x2])
    return stack(a, b, flatten=flatten)


import numpy as np
from pymoo.util.misc import stack


class MyTestProblem(MyProblem):

    def _calc_pareto_front(self, *args, **kwargs):
        return func_pf(**kwargs)

    def _calc_pareto_set(self, *args, **kwargs):
        return func_ps(**kwargs)


test_problem = MyTestProblem()


from pymoo.model.problem import FunctionalProblem


functional_test_problem = FunctionalProblem(2,
                                            objs,
                                            constr_ieq=constr_ieq,
                                            xl=-2,
                                            xu=2,
                                            func_pf=func_pf,
                                            func_ps=func_ps
                                            )


for problem in [vectorized_problem, elementwise_problem, functional_problem, test_problem, functional_test_problem]:
    from pymoo.algorithms.nsga2 import NSGA2
    from pymoo.factory import get_sampling, get_crossover, get_mutation


    print(problem.__class__.__name__)

    algorithm = NSGA2(
        pop_size=40,
        n_offsprings=10,
        sampling=get_sampling("real_random"),
        crossover=get_crossover("real_sbx", prob=0.9, eta=15),
        mutation=get_mutation("real_pm", eta=20),
        eliminate_duplicates=True
    )

    from pymoo.factory import get_termination

    termination = get_termination("n_gen", 40)

    from pymoo.optimize import minimize

    res = minimize(problem,
                   algorithm,
                   termination,
                   seed=1,
                   save_history=True,
                   verbose=True)

    if RESULT is None:
        RESULT = res
    else:
        np.testing.assert_allclose(RESULT.F, res.F)
        print("YES")

    from pymoo.visualization.scatter import Scatter

    # get the pareto-set and pareto-front for plotting
    ps = problem.pareto_set(use_cache=False, flatten=False)
    pf = problem.pareto_front(use_cache=False, flatten=False)

    # Design Space
    plot = Scatter(title="Design Space", axis_labels="x")
    plot.add(res.X, s=30, facecolors='none', edgecolors='r')
    if ps is not None:
        plot.add(ps, plot_type="line", color="black", alpha=0.7)
    plot.do()
    plot.apply(lambda ax: ax.set_xlim(-0.5, 1.5))
    plot.apply(lambda ax: ax.set_ylim(-2, 2))
    plot.show()

    # Objective Space
    plot = Scatter(title="Objective Space")
    plot.add(res.F)
    if pf is not None:
        plot.add(pf, plot_type="line", color="black", alpha=0.7)
    plot.show()

    n_evals = []  # corresponding number of function evaluations\
    F = []  # the objective space values in each generation
    cv = []  # constraint violation in each generation

    # iterate over the deepcopies of algorithms
    for algorithm in res.history:
        # store the number of function evaluations
        n_evals.append(algorithm.evaluator.n_eval)

        # retrieve the optimum from the algorithm
        opt = algorithm.opt

        # store the least contraint violation in this generation
        cv.append(opt.get("CV").min())

        # filter out only the feasible and append
        feas = np.where(opt.get("feasible"))[0]
        _F = opt.get("F")[feas]
        F.append(_F)

    import matplotlib.pyplot as plt

    k = min([i for i in range(len(cv)) if cv[i] <= 0])
    first_feas_evals = n_evals[k]
    print(f"First feasible solution found after {first_feas_evals} evaluations")

    plt.plot(n_evals, cv, '--', label="CV")
    plt.scatter(first_feas_evals, cv[k], color="red", label="First Feasible")
    plt.xlabel("Function Evaluations")
    plt.ylabel("Constraint Violation (CV)")
    plt.legend()
    plt.show()

    import matplotlib.pyplot as plt
    from pymoo.performance_indicator.hv import Hypervolume

    # MODIFY - this is problem dependend
    ref_point = np.array([1.0, 1.0])

    # create the performance indicator object with reference point
    metric = Hypervolume(ref_point=ref_point, normalize=False)

    # calculate for each generation the HV metric
    hv = [metric.calc(f) for f in F]

    # visualze the convergence curve
    plt.plot(n_evals, hv, '-o', markersize=4, linewidth=2)
    plt.title("Convergence")
    plt.xlabel("Function Evaluations")
    plt.ylabel("Hypervolume")
    plt.show()

    pf = problem.pareto_front(flatten=True, use_cache=False)

    import matplotlib.pyplot as plt
    from pymoo.performance_indicator.igd import IGD

    if pf is not None:
        # for this test problem no normalization for post prcessing is needed since similar scales
        normalize = False

        metric = IGD(pf=pf, normalize=normalize)

        # calculate for each generation the HV metric
        igd = [metric.calc(f) for f in F]

        # visualze the convergence curve
        plt.plot(n_evals, igd, '-o', markersize=4, linewidth=2, color="green")
        plt.yscale("log")  # enable log scale if desired
        plt.title("Convergence")
        plt.xlabel("Function Evaluations")
        plt.ylabel("IGD")
        plt.show()

    from pymoo.util.running_metric import RunningMetric

    running = RunningMetric(delta_gen=5,
                            n_plots=3,
                            only_if_n_plots=True,
                            key_press=False,
                            do_show=True)

    for algorithm in res.history[:15]:
        running.notify(algorithm)

    from pymoo.util.running_metric import RunningMetric

    running = RunningMetric(delta_gen=10,
                            n_plots=4,
                            only_if_n_plots=True,
                            key_press=False,
                            do_show=True)

    for algorithm in res.history:
        running.notify(algorithm)


from pymoo.algorithms.so_pso import PSO
from pymoo.factory import Rastrigin
from pymoo.optimize import minimize

problem = Rastrigin()

algorithm = PSO()

res = minimize(problem,
               algorithm,
               seed=1,
               verbose=False)

print("Best solution found: \nX = %s\nF = %s" % (res.X, res.F))


import matplotlib.pyplot as plt
import numpy as np

from pymoo.algorithms.so_genetic_algorithm import GA
from pymoo.model.repair import Repair
from pymoo.operators.crossover.order_crossover import OrderCrossover
from pymoo.operators.mutation.inversion_mutation import InversionMutation
from pymoo.operators.sampling.random_permutation_sampling import PermutationRandomSampling
from pymoo.optimize import minimize
from pymoo.problems.single.traveling_salesman import visualize, create_random_tsp_problem
from pymoo.util.termination.default import SingleObjectiveDefaultTermination
from pymoo.visualization.video.callback_video import AnimationCallback

problem = create_random_tsp_problem(50, 100, seed=1)


class StartFromZeroRepair(Repair):

    def _do(self, problem, pop, **kwargs):
        X = pop.get("X")
        I = np.where(X == 0)[1]

        for k in range(len(X)):
            i = I[k]
            x = X[k]
            _x = np.concatenate([x[i:], x[:i]])
            pop[k].set("X", _x)

        return pop


class PathVisualization(AnimationCallback):

    def notify(self, algorithm):
        if algorithm.n_gen % 10 == 0:

            fig, ax = plt.subplots()

            x = algorithm.opt[0].get("X")
            visualize(problem, x, fig=fig, ax=ax, show=False)
            ax.set_title(f"Generation: {algorithm.n_gen}")
            self.video.record(fig=fig)


algorithm = GA(
    pop_size=20,
    sampling=PermutationRandomSampling(),
    mutation=InversionMutation(),
    crossover=OrderCrossover(),
    repair=StartFromZeroRepair(),
    eliminate_duplicates=True
)

# if the algorithm did not improve the last 200 generations then it will terminate (and disable the max generations)
termination = SingleObjectiveDefaultTermination(n_last=200, n_max_gen=np.inf)

res = minimize(
    problem,
    algorithm,
    termination,
    # UNCOMMENT to save the visualization
    # callback=PathVisualization(fname="tsp.mp4"),
    verbose=False
)

print(res.F)
print(res.algorithm.evaluator.n_eval)

visualize(problem, res.X)




# START ctaea
from pymoo.algorithms.ctaea import CTAEA
from pymoo.factory import get_problem, get_reference_directions
from pymoo.optimize import minimize
from pymoo.visualization.scatter import Scatter

problem = get_problem("DASCMOP1", 2)

# create the reference directions to be used for the optimization
ref_dirs = get_reference_directions("das-dennis", problem.n_obj, n_points=91)

# create the algorithm object
algorithm = CTAEA(ref_dirs=ref_dirs)

# execute the optimization
res = minimize(problem,
               algorithm,
               ('n_gen', 600),
               seed=1,
               verbose=True
               )

plot = Scatter()
plot.add(problem.pareto_front(), plot_type="line", color="black", alpha=0.7)
plot.add(res.F, color="red")
plot.show()
# END ctaea

# START carside
problem = get_problem("carside")
ref_dirs = get_reference_directions("das-dennis", problem.n_obj, n_points=91)
algorithm = CTAEA(ref_dirs=ref_dirs)

res = minimize(problem,
               algorithm,
               ('n_gen', 600),
               seed=1,
               verbose=True
               )

Scatter().add(res.F).show()
# END carside



from pymoo.model.sampling import Sampling
import numpy as np


class PermutationRandomSampling(Sampling):
    """
    Generate random permutation.
    """
    def _do(self, problem, n_samples, **kwargs):
        X = np.full((n_samples, problem.n_var), 0, dtype=int)
        for i in range(n_samples):
            X[i, :] = np.random.permutation(problem.n_var)
        return X


from pymoo.model.population import Population
from pymoo.model.survival import Survival


class AgeBasedSurvival(Survival):

    def __init__(self, survival, n_max_age) -> None:
        super().__init__()
        self.survival = survival
        self.n_max_age = n_max_age

    def _do(self, problem, pop, n_survive, algorithm=None, **kwargs):

        if algorithm is None:
            raise Exception("Algorithm object needs to be passed to determine the current generation!")

        # get the generations of the population
        gen = pop.get("n_gen")

        # by default will with the current generation if unknown (should not happen)
        for k in range(len(gen)):
            if gen[k] is None:
                gen[k] = algorithm.n_gen

        # get the age of each individual
        age = algorithm.n_gen - gen

        # define what individuals are too old and which are young enough
        too_old = age > self.n_max_age
        young_enough = ~too_old

        # initialize the survivors
        survivors = Population()

        # do the survival with individuals being young enough
        if young_enough.sum() > 0:
            survivors = self.survival.do(problem, pop[young_enough], n_survive, algorithm=algorithm, **kwargs)

        n_remaining = n_survive - len(survivors)

        # if really necessary fill up with individuals which are actually too old
        if n_remaining > 0:
            fill_up = pop[too_old]
            fill_up = fill_up[fill_up.get("n_gen").argsort()]

            survivors = Population.merge(survivors, fill_up[:n_remaining])

        return survivors



from pymoo.algorithms.so_genetic_algorithm import GA
from pymoo.operators.crossover.order_crossover import OrderCrossover
from pymoo.operators.mutation.inversion_mutation import InversionMutation
from pymoo.operators.sampling.random_permutation_sampling import PermutationRandomSampling
from pymoo.optimize import minimize
from pymoo.problems.single.flowshop_scheduling import visualize, create_random_flowshop_problem
from pymoo.util.termination.default import SingleObjectiveDefaultTermination


problem = create_random_flowshop_problem(n_machines=5, n_jobs=10, seed=1)

algorithm = GA(
    pop_size=20,
    eliminate_duplicates=True,
    sampling=PermutationRandomSampling(),
    mutation=InversionMutation(),
    crossover=OrderCrossover()
)

# if the algorithm did not improve the last 200 generations then it will terminate
termination = SingleObjectiveDefaultTermination(n_last=50, n_max_gen=10000)


res = minimize(
    problem,
    algorithm,
    termination,
    seed=1,
    verbose=True
)

visualize(problem, res.X)



from pymoo.algorithms.nsga3 import NSGA3
from pymoo.factory import get_reference_directions
from pymoo.optimize import minimize
from pymoo.problems.multi.sympart import SYMPART, SYMPARTRotated
from pymoo.visualization.scatter import Scatter
import matplotlib.pyplot as plt


for problem, name in zip([SYMPART(), SYMPARTRotated()], ["SYM-PART", "SYM-PART rotated"]):

    ref_dirs = get_reference_directions("das-dennis", problem.n_obj, n_partitions=20)
    PS = problem.pareto_set(500)
    PF = problem.pareto_front(500)

    algorithm = NSGA3(ref_dirs=ref_dirs)

    res = minimize(problem,
                   algorithm,
                   ('n_gen', 500),
                   seed=1,
                   verbose=False)

    fig_name = f"{algorithm.__class__.__name__} on {name}"
    # visualize decision space
    plot = Scatter(title="Decision Space")
    plot.add(PS, s=10, color='r', label="PS")
    plot.add(res.X, s=30, color='b', label="Obtained solutions")
    plot.do()
    plt.legend()

    # visualize objective space
    plot = Scatter(title="Objective Space")
    plot.add(PF, s=10, color='r', label="PF")
    plot.add(res.F, s=30, color='b', label="Obtained solutions")
    plot.do()
    plt.legend()

    plt.show()



from timeit import timeit
# noinspection PyUnresolvedReferences
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting

import numpy as np

# generate random data samples
F = np.random.random((1000, 2))

# use fast non-dominated sorting
res = timeit("NonDominatedSorting(method=\"fast_non_dominated_sort\").do(F)", number=10, globals=globals())
print(f"Fast ND sort takes {res} seconds")

# # use efficient non-dominated sorting with sequential search, this is the default method
# res = timeit("NonDominatedSorting(method=\"efficient_non_dominated_sort\").do(F, strategy=\"sequential\")", number=10,
#              globals=globals())
# print(f"Efficient ND sort with sequential search (ENS-SS) takes {res} seconds")
#
#
# res = timeit("NonDominatedSorting(method=\"efficient_non_dominated_sort\").do(F, strategy=\"binary\")", number=10,
#              globals=globals())
# print(f"Efficient ND sort with binary search (ENS-BS) takes {res} seconds")



from abc import abstractmethod

import numpy as np

from pymoo.model.population import Population
from pymoo.util.clearing import select_by_clearing, func_select_from_sorted
from pymoo.util.misc import norm_eucl_dist


# ---------------------------------------------------------------------------------------------------------
# Constraint Handling
# ---------------------------------------------------------------------------------------------------------


class LeastInfeasibleSurvival:

    def __init__(self, clearing=True, clearing_delta=0.05) -> None:
        super().__init__()
        self.clearing = clearing
        self.clearing_delta = clearing_delta

    def do(self, problem, pop, n_survive, **kwargs):

        # sort them by cv
        sorted_by_cv = pop.get("CV")[:, 0].argsort()
        pop = pop[sorted_by_cv]

        if self.clearing:

            # calculate the distance from each individual to another - pre-processing for the clearing
            X = pop.get("X")
            D = norm_eucl_dist(problem, X, X)

            # select by clearing using the order defined before
            I = select_by_clearing(pop, D, n_survive, func_select_from_sorted, delta=self.clearing_delta)

            return pop[I]

        else:
            return pop[:n_survive]


# ---------------------------------------------------------------------------------------------------------
# Survival
# ---------------------------------------------------------------------------------------------------------


class Survival:

    def __init__(self,
                 filter_infeasible=True,
                 infeas_survival=LeastInfeasibleSurvival(clearing=False)):
        """
        The survival process is implemented inheriting from this class, which selects from a population only
        specific individuals to survive. This base class can take care of splitting the feasible and infeasible
        solutions first. By default infeasible solutions are simply sorted by their constraint violation.


        Parameters
        ----------

        filter_infeasible : bool
            Whether for the survival infeasible solutions should be filtered first

        """
        super().__init__()
        self.filter_infeasible = filter_infeasible
        self.infeas_survival = infeas_survival

    def do(self,
           problem,
           pop,
           n_survive,
           n_min_infeas_survive=0,
           return_indices=False,
           **kwargs):

        # make sure the population has at least one individual
        if len(pop) == 0:
            return pop

        n_survive = min(n_survive, len(pop))

        # if the split should be done beforehand
        if self.filter_infeasible and problem.n_constr > 0:

            # initialize the feasible and infeasible population
            feas_pop, infeas_pop = Population(), Population()

            # split feasible and infeasible solutions
            feas, infeas = split_by_feasibility(pop, sort_infeasbible_by_cv=True)

            # if there was no feasible solution was added at all - which means at least one infeasible
            if len(feas) == 0:
                infeas_pop = self.infeas_survival.do(problem, pop[infeas], n_survive, **kwargs)

            # if there are feasible solutions in the population
            else:

                # if feasible solutions do exist
                if len(feas) > 0:
                    feas_pop = self._do(problem, pop[feas], min(len(feas), n_survive), **kwargs)

                # calculate how many individuals are still remaining to be filled up with infeasible ones
                n_remaining = n_survive - len(feas_pop)

                # the maximum of necessary to survive or minimum to be defined being infeasible
                n_infeas_survive = max(n_remaining, n_min_infeas_survive)

                # no more than all infeasible solutions can survive
                n_infeas_survive = min(len(infeas), n_infeas_survive)

                # if infeasible solutions needs to be added
                if n_infeas_survive > 0:
                    infeas_pop = self.infeas_survival.do(problem, pop[infeas], n_infeas_survive, **kwargs)

            survivors = Population.merge(feas_pop, infeas_pop)

        else:
            survivors = self._do(problem, pop, n_survive, **kwargs)

        if return_indices:
            H = {}
            for k, ind in enumerate(pop):
                H[ind] = k
            return [H[survivor] for survivor in survivors]
        else:
            return survivors

    @abstractmethod
    def _do(self, problem, pop, n_survive, **kwargs):
        pass


def split_by_feasibility(pop, sort_infeasbible_by_cv=True):
    CV = pop.get("CV")

    b = (CV <= 0)

    feasible = np.where(b)[0]
    infeasible = np.where(np.logical_not(b))[0]

    if sort_infeasbible_by_cv:
        infeasible = infeasible[np.argsort(CV[infeasible, 0])]

    return feasible, infeasible



import numpy as np

from pymoo.interface import crossover
from pymoo.operators.crossover.parent_centric_crossover import PCX
from pymoo.visualization.scatter import Scatter

X = np.eye(3)
X[1] = [0.9, 0.1, 0.1]

n_points = 1000

a = X[[0]].repeat(n_points, axis=0)
b = X[[1]].repeat(n_points, axis=0)
c = X[[2]].repeat(n_points, axis=0)

obj = PCX(eta=0.1, zeta=0.1, impl="elementwise")

_X = crossover(obj, a, c, b, xl=-1, xu=1)
sc = Scatter()
sc.add(_X, facecolor=None, edgecolor="blue", alpha=0.7)
sc.add(X, s=100, color="red")
sc.add(X.mean(axis=0), color="green", label="Centroid")
sc.show()



import numpy as np

from pymoo.algorithms.rnsga2 import RNSGA2
from pymoo.factory import get_problem
from pymoo.model.callback import Callback
from pymoo.optimize import minimize
from pymoo.performance_indicator.rmetric import RMetric

problem = get_problem("zdt1", n_var=30)
pf = problem.pareto_front()
ref_points = np.array([[0.5, 0.2], [0.1, 0.6]])


class MyCallback(Callback):

    def notify(self, algorithm):
        rmetric = RMetric(algorithm.problem, ref_points)
        rigd, rhv = rmetric.calc(algorithm.opt.get("F"))
        print(f"R-IGD: {rigd}, R-HV: {rhv}")


algorithm = RNSGA2(
    ref_points=ref_points,
    pop_size=40,
    epsilon=0.01,
    normalization='front',
    extreme_points_as_reference_points=False,
    weights=np.array([0.5, 0.5]))

res = minimize(problem,
               algorithm,
               save_history=True,
               callback=MyCallback(),
               termination=('n_gen', 300),
               seed=1,
               pf=pf,
               verbose=False)

rmetric = RMetric(problem, ref_points,  delta=0.2)
rigd, rhv = rmetric.calc(res.F, others=None)
print(rigd, rhv)
